require "vizshot_gfdnavi"
require "pp"

Dependencies.require_or_load "execute_analysis"

class KnowledgeController < ApplicationController
  include ExecuteAnalysis

  layout "gfdnavi", :except => [:list_without_layout, :show_without_layout]
  before_filter :login_required, :except => [:index, :list, :list_without_layout, :show, :show_without_layout, :fig2analysis]

  def index
    list
    render :action => 'list'
  end

  verify :method => :post, :only => [ :destroy, :create, :update ], :redirect_to => { :action => :list }

  # paginate𗘗pĒmhLg̃Xg\
  def list
    if Knowledge.respond_to?("paginate")
      # for Rails 2.*
      user = (login=session[:user]) && User.find_by_login(login)
      @knowledges = Knowledge.paginate(:all, {:user => user}, {:page => params[:page], :per_page => 5, :order=>"mtime DESC"})
    else
      # for Rails 1.*
      @knowledge_pages, @knowledges = paginate :knowledges, :per_page => 5
      @knowledge = Knowledge.find(:all, :user => user, :order =>"mtime DESC")
    end
  end

  # ʏ㕔Gfdnavĩj[olist\
  def list_without_layout
    # if params[:knowledge_ids], list_without_layout
    knowledge_ids = params[:knowledge_ids].collect! {|id| id.to_i}
    node_ids = Array.new
    for kid in knowledge_ids
      node_ids.push(Knowledge.find_by_id(kid).node.id)
    end

    user = (login=session[:user]) && User.find_by_login(login)
    if Knowledge.respond_to?("paginate")
      # for Rails 2.*
      @knowledges = Knowledge.paginate(:all, {:user => user}, {:page => params[:page], :per_page => 5, :conditions => ["(node_id IN (?))", node_ids], :order =>"mtime DESC"})
    else
      # for Rails 1.*
      @knowledge_pages, @knowledges = paginate :knowledges, :per_page => 5
      @knowledges = []
      for nid in node_ids
        @knowledges.push(Knowledge.find(user, :first, {:conditions => ["id = ?", nid]}))
      end
      @knowledges = @knowledges.uniq
      @knowledges.sort! {|a,b| b.mtime <=> a.mtime}
    end
    render :action => 'list'
  end

  # mhLg̉{
  def show
    begin
      user = (login=session[:user]) && User.find_by_login(login)
      @knowledge = Knowledge.find(:first, :conditions => ["path=?", params[:path]], :user => user)
      @knowledge_figures = KnowledgeFigure.find(:all, :conditions => ["knowledge_id=?", @knowledge.id]) # not find -> return []
    rescue
      flash[:notice] = "The knowledge document doesn't exist."
      redirect_to :action => 'list'
    else
      # = G̃pXƑ傫
      #   * G̃pXnodese[upath擾
      #   * G̑傫imagese[uvizshot擾
      @image_paths = Array.new
      @image_widths = Array.new
      @image_heights = Array.new
      @caption_widths = Array.new
      @caption_heights = Array.new
      @knowledge_figures.each do |@kf|
        unless (image = Image.find_by_id(@kf.image_id))
          raise "Figure is not found.\n"
        else
          @image_paths << image.path
          vizshot_yaml = image.vizshot
          vizshot = YAML.load(vizshot_yaml)
          if (image_size = vizshot.get_size)
            @image_widths << image_size[0]
            @image_heights << image_size[1]
            @caption_widths << (image_size[0] * 0.8).to_i
            @caption_heights << (image_size[1] * 0.8).to_i
          else
            # vizshot畜łȂƂnormalTCYKp
            @image_widths << 554
            @image_heights << 554
            @caption_widths << 440
            @caption_heights << 440
          end
        end
      end
    end
  end

  # ʏ㕔Gfdnavĩj[olists
  def show_without_layout
    show
    render :action => 'show'
  end

  # list.rhtml -> new.rhtml (Click "New Knowledge".)
  def new
    @type = "new"
    @knowledge = Knowledge.new
    @user = (login=session[:user]) && User.find_by_login(login)
    @groups = @user.belonging_groups
    @num_of_figure = 1
  end

  # Click "Click a Knowledge Document with this/these Image(s)" button in Analysis
  def new_from_analysis
#    print "((((((((((( new_from_analysis ))))))))))))))\n"
    session[:diagrams].each do |dia|
=begin
      print "id: "
      pp DiagramCache.find(dia)[:id]
      print "path: "
      pp DiagramCache.find(dia)[:path]
      print "DiagramCacheDatum = "
      pp DiagramCacheDatum.find_by_diagram_cache_id(DiagramCache.find(dia)[:id])
=end
    end

    
    
    @type = "new_from_analysis"
    @knowledge = Knowledge.new
    @user = (login=session[:user]) && User.find_by_login(login)
    @groups = @user.belonging_groups
    session[:knowledge] = Hash.new
    
    # ʓɂG̔ԍ̔z擾
    if (diagram_ids = session[:diagrams].dup)
      session[:knowledge_figure] = Hash.new
      session[:knowledge_figure]["diagram_ids"] = diagram_ids
    else
      raise "diagram is not found."
    end
    @num_of_figure = diagram_ids ? diagram_ids.length : 0
  end

  # Edit a knowledge document.
  # title, textbody, default_layout ̓tH[ɎœB
  def edit
    @type = "edit"
    @user = (login=session[:user]) && User.find_by_login(login)
    @groups = @user.belonging_groups
    
    # list.rhtml -> edit.rhtml (Click "Edit".)
    if params["from_list"] == "from_list"
      @knowledge = Knowledge.find(:first, :conditions => ["path=?", params[:path]], :user => @user)
      session[:knowledge] = Hash.new
      session[:knowledge_figure] = Hash.new
      @figure_captions = Array.new
      @figure_paths = Array.new
      
=begin
      print "@knowledge = "
      pp @knowledge
      print "@knowledge.title = "
      pp @knowledge.title
      print "@knowledge.methods = "
      pp @knowledge.methods.sort
      print "@knowledge.knowledges.title = "
      pp @knowledge.knowledges.title
=end

      knowledge_figures = @knowledge.knowledge_figures
      knowledge_figures.each do |kf|
        img = Image.find_by_id(kf.image_id) # ꎞۑ̂Ƃ́Aimage_idnullɂȂBgĕ򂳂΂B
        @figure_captions << kf.caption
        @figure_paths << img.node.path
      end
    end

    @num_of_figure = @knowledge.knowledge_figures.length

  end

  def destroy
    user = (login=session[:user]) && User.find_by_login(login)
    @knowledge = Knowledge.find(:first, :conditions => ["path=?", params[:path]], :user => user)

    # .knlget@C̏
    File.delete(@knowledge.fname)

    # f[^x[X̏
    @knowledge.destroy
    redirect_to :action => 'list'
  end

  # create a knowledge document in new.rhtml
  def create_from_new
    if params[:commit] == "Create"
      create
    else
      redirect_to :action => "new"
    end
  end

  # create a knowledge document in edit.rhtml
  def create_from_edit
    if (params[:commit] == "Save" || params[:commit] == "Save As")
      create
    else
      redirect_to :action => "edit", :id => params[:id]
    end
  end

  # == create a knowledge document in Analysis.
  # Save images in this method, and call create method to make a knowledge document.
  def create_from_analysis
    user = (login=session[:user]) && User.find_by_login(login)
    unless user
      redirect_to :controller=> "knowledge", :action => "index"
      return
    end

    unless params[:commit] == "Create"
      redirect_to :action => "new_from_analysis"
    else
      ActiveRecord::Base::transaction do
        # === save images
        dirs = params[:node]["path"].split(File::Separator)[0..-2]
        dirname = dirs.join("/")
        dirname = dirname + "/" unless dirname == ""
        filename = params[:node]["path"].split(File::Separator)[-1]
        
#        print "params[knowledge_figure] = "
#        pp params[:knowledge_figure]
        
        if params[:knowledge_figure]
          params[:knowledge_figure].each do |key, value|
=begin
            print "------\n"
            print "key = "
            pp key
            print "value = "
            pp value
=end
            case key.to_i
            when  0..9  then auto_figure_path = "image00" + key.to_s
            when 10..99 then auto_figure_path = "image0"  + key.to_s
            else             auto_figure_path = "image" + key.to_s
            end
            id = value["diagram_id"]
            type = "diagram"
            name = value["filename"]
            path = "knowledge/" + dirname + "." + filename + "/" + auto_figure_path
            group = params[:knowledge]["group_name"]
            
            # determine realpath for figure.
            value["figure_path"] = "/usr/" + user.login + "/" + path
            
            # about groups
            unless params[:groups]
              raise "Check at least 1 group to visible!\n"
            else
              groups = params[:groups].values # Hash to Array
            end
            if groups[0] == 'everyone'
              other_mode = 4
              groups.shift
            else
              other_mode = 0
            end
            
            refs = Array.new
            var = Image.new
            diagram = get_temp_diagram(id)
            viz = YAML.load(diagram.vizshot)
            ids = viz.get_variables.collect{|v|    # ̃ubN͂Ƃ肠̂܂܂ɂĂ邪AvΉ
              if Variable===v && ! v.id
                from = v.fname
                to, pa = auto_file_path("data")
                File.move(from, to)
                v.path = pa + '/' + v.name
                v.owner = user
                v.other_mode = other_mode
                v.set_rgroups(groups) if groups.length > 0
                v.save!
              else
                v = Variable===v ? v : Variable.find(:first, :conditions=>["path=?",v],:user=>user)
              end
              refs << v.node
              v.id
            }
            var.vizshot = viz.to_yaml
            path = "/#{path}.png"
            # from = diagram.path
            from = value["diagram_path"]
            name = var.name = File.basename(path)
            var.path = File.join("/usr",user.login,path)

            var.owner = user
            var.other_mode = other_mode
            var.node.set_rgroups(groups) if groups.length > 0
            
            # m쐬Ɠɕۑ摜ɂkeyword_attributes͂Ȃ
=begin
            var.keyword_attributes.build(:name => "description", :value => description) unless description == ""
            keywords.each{|k,v|
              if Hash === v && (name=v["name"])!="" && (val=v["value"])!=""
                if val.to_i.to_s == val
                  val = NArray[val.to_i]
                elsif val.to_f.to_s == val
                  val = NArray[val.to_f]
                end
                var.keyword_attributes.build(:name => name, :value => val)
              end
            }
=end

            to = var.fname
            File.makedirs( File.dirname(to) )
            
            ## If none, make directory.
            # make directories
            make_directories(var, user, groups, other_mode)
            # save images
            if var.save
              rtype = {"diagram"=>"draw", "data"=>"analyze"}[type]
              refs.each{|ref|
                NodeRelation.new(:name=>rtype, :reference=>ref, :referenced_by=>var.node).save!
              }
              File.copy( from, to )
            else
              flash[:notice] = "Failed to save.\n"
              redirect_to :action => "list"
            end
          end
        end
        create
      end
    end
  end

  # create and save Knowledge Document
  def create
    # == knowledges table
    # === Click "Save" button in edit.rhtml
    # if "Save", insert all original data include "@Knowledge.id"
    user = (login=session[:user]) && User.find_by_login(login)
    if params[:commit]=="Save"
      unless @knowledge = Knowledge.find(:first, :conditions => ["path=?", params[:original_path]], :user => user)
        raise "Specified knowledge is not found or not yours."
      end
    else
      @knowledge = Knowledge.new
    end

    # about groups
    raise "Check at least 1 group to visible!\n" unless params[:groups]
    groups = params[:groups].values # Hash to Array

    if groups[0] == 'everyone'
      other_mode = 4
      groups.shift
    else
      other_mode = 0
    end
    @knowledge.other_mode = other_mode

    # insert title, textbody(and change for RD-like format)
    # , default_layout, horizontal number of figures
    # , height/width, %/px, and size(number).
    horizontal_figures = params[:knowledge]["horizontal_figures"] || 1
    figures_size_status = params[:knowledge]["height_or_width"] == "width" ? 1 : 0
    figures_size_status += 2 if params[:knowledge]["percent_or_pixel"] == "pixel"
    
    @knowledge.attributes = {
      "title" => params[:knowledge]["title"],
      "author" => params[:knowledge]["author"],
      "textbody" => params[:knowledge]["textbody"], 
      "default_layout" => params[:knowledge]["default_layout"], 
      "horizontal_figures" => horizontal_figures,
      "figures_size_status" => figures_size_status,
      "figures_size_number" => params[:knowledge]["figures_size_number"]}

    # == nodes table
    # insert name, path, owner for knowledge.node
    @knowledge.name = params[:knowledge]["title"]
    if params[:original_path] && params[:commit] == "Save"
      @knowledge.path = params[:original_path]
    elsif params[:node]
      @knowledge.path = "/usr/"+ session[:user] + "/knowledge/" + params[:node]["path"] + ".knlge"
    else
      if params[:commit]=="Save"
        @knowledge.path = params[:original_path]
      else
        @knowledge.path = params[:new_path]
      end
    end
    @knowledge.owner = user
    @knowledge.node.set_rgroups(groups) if groups.length > 0

    # === If none, make directory.
    # * path in the disc
    File.makedirs(File.dirname(@knowledge.fname)) or raise("failed to makedir")
    # * parh in gfdnavi
    make_directories(@knowledge, user, groups, other_mode)

    # == knowledge_figures table
    # if "Save", at first keep original knowledge_figures and keyword attributes that are already in the database. (delete them later.)
    if params[:commit]=="Save"
      old_knowledge_figures = KnowledgeFigure.find(:all, :conditions => ["knowledge_id = ?", params[:id]])
      old_keyword_attributes = KeywordAttribute.find(:all, :conditions => ["node_id = ?", @knowledge.node])
    end

    # paramsœnꂽGArray(knowledgefigures)Ɋi[
    knowledgefigures = Array.new
    if params[:knowledge_figure]
      params[:knowledge_figure].sort {|a, b| a[0].to_i <=> b[0].to_i}.each do |key, value| # G̏Ԃ𐳂\[g
        if value["figure_path"] != ""
          value["figure_path"] = value["figure_path"] + ".png"
          unless img = Image.find(:first, :conditions => ["path=?", value["figure_path"]], :user => user)
            raise "Figure not found."
          end
          knowledgefigure = KnowledgeFigure.new(:caption => value["figure_caption"], :image_id => img.id)
          knowledgefigure.knowledge = @knowledge
          knowledgefigures.push(knowledgefigure)
        end
      end
    end

    # ۑ悤ƂĂknowledgeIuWFNgɊG֘At
    @knowledge.knowledge_figures = knowledgefigures
    
    # pɁAtitle  textbody  keyword_attributes ɓB
    @knowledge.keyword_attributes.build(:name => "title", :value => params[:knowledge]["title"])
    @knowledge.keyword_attributes.build(:name => "textbody", :value => params[:knowledge]["textbody"])
    
    # XVAmtime㏑
    @knowledge.mtime = Time.new unless params[:commit] == "Create"

    # save
    if @knowledge.save
      flash[:notice] = "A knowledge document is successfully saved."

      # if "Save", delete original knowledge_figures.
      if params[:commit] == "Save"
        old_knowledge_figures.each do |old_knowledge_figure|
          old_knowledge_figure.destroy
        end
        old_keyword_attributes.each do |old_keyword_attribute|
          old_keyword_attribute.destroy
        end
      end

      # == .knlge t@C쐬
      # * knowledge e[u̗vf
      knowledge_hash = {
        "gfdnavi_knowledge" => 
        {
          "title" => @knowledge.title,
          "author" => @knowledge.author, 
          "textbody" => @knowledge.textbody,
          "default_layout" => @knowledge.default_layout,
          "horizontal_figures"=> @knowledge.horizontal_figures,
          "knowledge_figures" => {},
          "owner" => @knowledge.owner.login,
          "figures_size_status" => @knowledge.figures_size_status,
          "figures_size_number" => @knowledge.figures_size_number
        }}
      # * knowledge_figures e[u̗vf
      @knowledge.knowledge_figures.length.times do |i|
        knowledge_hash["gfdnavi_knowledge"]["knowledge_figures"].update({ "#{i}" => {"image_id" => @knowledge.knowledge_figures[i].image_id, "caption"  => @knowledge.knowledge_figures[i].caption}})
      end
      # .knlget@Cւ̏
      File.open(@knowledge.fname,"w"){|file| file.print knowledge_hash.to_yaml}
      
      redirect_to :action => 'show', :path => @knowledge.path
    else
      # error.
      if params[:knowledge]["title"] == ""
        flash[:notice] = "Title is empty. Failed to save."
        redirect_to :action => "list"
      else
        flash[:notice] = "Failed to save."
        redirect_to :action => 'list'
      end
    end
  end


  # view  _layout.rhtml Ŏgp
  def fig2analysis
    vizshot_yaml = Image.find_by_id(params[:image_id]).vizshot
    vizshot = YAML.load(vizshot_yaml)
    
    analysis = Analysis.from_vizshot(vizshot)

    session[:analysis] = analysis
=begin
    print "analysis = "
    pp analysis
    print "analysis.attributes[variables] = "
    pp analysis.attributes["variables"]
    print "Node.find_by_id(analysis.attributes[variables].node_id) = "
    pp Node.find_by_id(analysis.attributes["variables"][0].attributes["node_id"]).path

    print "################knowledge fig2analysis##################\n"
    print "session[analysis] = "
    pp session[:analysis]
    print "################knowledge fig2analysis##################\n"
=end
    variable = Array.new
    variable = [analysis.attributes["variables"], Node.find_by_id(analysis.attributes["variables"][0].attributes["node_id"]).path]
    
    session[:variables_list] = variable
    
    #session[variables_list] = [#<Variable id: 9, node_id: 26>, "/samples/reanalysis/era40/t.jan.nc/t"]
    #Kv

    redirect_to :controller => "analysis", :action => "index"
  end

private
    def get_temp_diagram(id)
    diagrams = session[:diagrams]
    if diagrams
      /^temp_(.*)/ =~ id
      id = $1.to_i
      return DiagramCache.find(diagrams[id])
    else
      return nil
    end
  end

private
  def make_directories(obj, user, groups, other_mode)
    full_path = ""
    parent = nil

    obj.path.split(File::Separator)[0..-2].each{|dname|
      full_path = File.join(full_path, dname)
      dir = Directory.find(:first, :conditions=>["path=?",full_path], :user=>user)
      unless dir
        dir = Directory.new
        dir.name = dname
        dir.path = full_path
        dir.parent = parent.node
        dir.owner = user
        dir.other_mode = other_mode
        dir.node.set_rgroups(groups) if groups.length > 0
        if full_path == obj.file
          dir.downloadable = obj.downloadable?
          dir.plain_file = true
        end
        dir.save!
      end
      parent = dir
    }
  end
end
